# -*- coding: utf-8 -*-
import mock
import pytest

from s4 import sync
from s4.clients import SyncState, local, s3
from s4.sync import Resolution

from tests import utils


class TestGetStates(object):
    def test_empty_clients(self, s3_client, local_client):
        worker = sync.SyncWorker(s3_client, local_client)
        actual_output = list(worker.get_states())
        assert actual_output == []


class TestGetSyncStates(object):
    def test_empty(self, local_client, s3_client):
        assert sync.SyncWorker(local_client, s3_client).get_sync_states() == ({}, {})

    def test_correct_output(self, local_client, s3_client):
        utils.set_local_index(
            local_client,
            {
                "chemistry.txt": {"local_timestamp": 9431, "remote_timestamp": 9431},
                "physics.txt": {"local_timestamp": 10000, "remote_timestamp": 10000},
                "maltese.txt": {"local_timestamp": 7000, "remote_timestamp": 6000},
            },
        )
        utils.set_s3_index(
            s3_client,
            {
                "chemistry.txt": {"local_timestamp": 10000, "remote_timestamp": 9431},
                "physics.txt": {"local_timestamp": 13000, "remote_timestamp": 12000},
                "maltese.txt": {"local_timestamp": 6000, "remote_timestamp": 6000},
            },
        )

        utils.set_local_contents(local_client, "history.txt", timestamp=5000)
        utils.set_s3_contents(s3_client, "art.txt", timestamp=200000)
        utils.set_local_contents(local_client, "english.txt", timestamp=90000)
        utils.set_s3_contents(s3_client, "english.txt", timestamp=93000)
        utils.set_s3_contents(s3_client, "chemistry.txt", timestamp=10000)
        utils.set_local_contents(local_client, "physics.txt", timestamp=11000)
        utils.set_s3_contents(s3_client, "physics.txt", timestamp=13000)
        utils.set_local_contents(local_client, "maltese.txt", timestamp=7000)
        utils.set_s3_contents(s3_client, "maltese.txt", timestamp=8000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            "english.txt": (
                SyncState(SyncState.CREATED, 90000, None),
                SyncState(SyncState.CREATED, 93000, None),
            ),
            "physics.txt": (
                SyncState(SyncState.UPDATED, 11000, 10000),
                SyncState(SyncState.NOCHANGES, 13000, 12000),
            ),
        }
        expected_resolutions = {
            "maltese.txt": Resolution(
                Resolution.UPDATE, local_client, s3_client, "maltese.txt", 8000
            ),
            "chemistry.txt": Resolution(
                Resolution.DELETE, s3_client, None, "chemistry.txt", 9431
            ),
            "history.txt": Resolution(
                Resolution.CREATE, s3_client, local_client, "history.txt", 5000
            ),
            "art.txt": Resolution(
                Resolution.CREATE, local_client, s3_client, "art.txt", 200000
            ),
        }
        assert unhandled_events == expected_unhandled_events
        assert resolutions == expected_resolutions

    def test_nochanges_but_different_remote_timestamps(self, local_client, s3_client):
        utils.set_local_index(
            local_client,
            {"german.txt": {"local_timestamp": 4000, "remote_timestamp": 4000}},
        )
        utils.set_s3_index(
            s3_client,
            {"german.txt": {"local_timestamp": 6000, "remote_timestamp": 6000}},
        )
        utils.set_local_contents(local_client, "german.txt", timestamp=4000)
        utils.set_s3_contents(s3_client, "german.txt", timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_resolutions = {
            "german.txt": Resolution(
                "UPDATE", local_client, s3_client, "german.txt", 6000
            )
        }
        assert resolutions == expected_resolutions
        assert unhandled_events == {}

    def test_updated_but_different_remote_timestamp(self, local_client, s3_client):
        utils.set_local_index(
            local_client,
            {"biology.txt": {"local_timestamp": 4000, "remote_timestamp": 3000}},
        )
        utils.set_s3_index(
            s3_client,
            {"biology.txt": {"local_timestamp": 6000, "remote_timestamp": 6000}},
        )
        utils.set_local_contents(local_client, "biology.txt", timestamp=4500)
        utils.set_s3_contents(s3_client, "biology.txt", timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            "biology.txt": (
                SyncState(SyncState.UPDATED, 4500, 3000),
                SyncState(SyncState.NOCHANGES, 6000, 6000),
            )
        }
        assert resolutions == {}
        assert unhandled_events == expected_unhandled_events

    def test_deleted_but_different_remote_timestamp(self, local_client, s3_client):
        utils.set_local_index(
            local_client,
            {"chemistry.txt": {"local_timestamp": 4000, "remote_timestamp": 3000}},
        )
        utils.set_s3_index(
            s3_client,
            {"chemistry.txt": {"local_timestamp": 6000, "remote_timestamp": 6000}},
        )
        utils.set_s3_contents(s3_client, "chemistry.txt", timestamp=6000)

        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        expected_unhandled_events = {
            "chemistry.txt": (
                SyncState(SyncState.DELETED, None, 3000),
                SyncState(SyncState.NOCHANGES, 6000, 6000),
            )
        }
        assert resolutions == {}
        assert unhandled_events == expected_unhandled_events

    def test_deleted_doesnotexist(self, local_client, s3_client):
        utils.set_local_index(
            local_client,
            {"physics.txt": {"local_timestamp": 5000, "remote_timestamp": 4550}},
        )
        worker = sync.SyncWorker(local_client, s3_client)
        resolutions, unhandled_events = worker.get_sync_states()
        assert resolutions == {}
        assert unhandled_events == {}


def assert_contents(clients, key, data=None):
    for client in clients:
        sync_object = client.get(key)
        if data is not None:
            assert sync_object.fp.read() == data


def assert_remote_timestamp(clients, key, expected_timestamp):
    for client in clients:
        assert client.get_remote_timestamp(key) == expected_timestamp


def assert_local_keys(clients, expected_keys):
    for client in clients:
        assert sorted(client.get_local_keys()) == sorted(expected_keys)


def assert_existence(clients, keys, exists):
    for client in clients:
        for key in keys:
            assert (client.get(key) is not None) == exists


class TestSyncWorker(object):
    def test_repr(self):
        local_client = local.LocalSyncClient("/home/bobs/burgers")
        s3_client = s3.S3SyncClient(None, "burgerbucket", "foozie")
        worker = sync.SyncWorker(local_client, s3_client)
        assert (
            repr(worker) == "SyncWorker</home/bobs/burgers/, s3://burgerbucket/foozie/>"
        )


class TestIntegrations(object):
    def test_dry_run(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, "colors/cream", 9999, "#ddeeff")

        utils.set_local_contents(local_client, "colors/red", 5000, "#ff0000")
        utils.set_local_contents(local_client, "colors/green", 3000, "#00ff00")
        utils.set_local_contents(local_client, "colors/blue", 2000, "#0000ff")

        local_client.put = mock.MagicMock()
        s3_client.put = mock.MagicMock()

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(dry_run=True)

        assert not local_client.put.called
        assert not s3_client.put.called

    def test_action_callback(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, "colors/cream", 9999)

        utils.set_local_contents(local_client, "colors/green", 3000)
        utils.set_local_contents(local_client, "colors/blue", 2000)

        callback_mock = mock.MagicMock()

        worker = sync.SyncWorker(local_client, s3_client, action_callback=callback_mock)
        worker.sync()

        assert callback_mock.call_count == 3
        assert callback_mock.call_args_list[0][0][0].key == "colors/blue"
        assert callback_mock.call_args_list[1][0][0].key == "colors/cream"
        assert callback_mock.call_args_list[2][0][0].key == "colors/green"

    def test_local_with_s3(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, "colors/cream", 9999, "#ddeeff")

        utils.set_local_contents(local_client, "colors/red", 5000, "#ff0000")
        utils.set_local_contents(local_client, "colors/green", 3000, "#00ff00")
        utils.set_local_contents(local_client, "colors/blue", 2000, "#0000ff")

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        expected_keys = ["colors/red", "colors/green", "colors/blue", "colors/cream"]
        assert_local_keys(clients, expected_keys)
        assert_contents(clients, "colors/red", b"#ff0000")
        assert_contents(clients, "colors/green", b"#00ff00")
        assert_contents(clients, "colors/blue", b"#0000ff")
        assert_contents(clients, "colors/cream", b"#ddeeff")
        assert_remote_timestamp(clients, "colors/red", 5000)
        assert_remote_timestamp(clients, "colors/green", 3000)
        assert_remote_timestamp(clients, "colors/blue", 2000)
        assert_remote_timestamp(clients, "colors/cream", 9999)

        utils.delete_local(local_client, "colors/red")

        worker.sync()
        expected_keys = ["colors/green", "colors/blue", "colors/cream"]
        assert_local_keys(clients, expected_keys)

    def test_fresh_sync(self, local_client, s3_client):
        utils.set_local_contents(local_client, "foo", timestamp=1000)
        utils.set_local_contents(local_client, "bar", timestamp=2000)
        utils.set_s3_contents(s3_client, "baz", timestamp=3000, data="what is up?")

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        assert_local_keys(clients, ["foo", "bar", "baz"])
        assert_remote_timestamp(clients, "foo", 1000)
        assert_remote_timestamp(clients, "bar", 2000)
        assert_remote_timestamp(clients, "baz", 3000)
        assert_existence(clients, ["foo", "bar", "baz"], True)
        assert_contents(clients, "baz", b"what is up?")

        utils.delete_local(local_client, "foo")
        utils.set_local_contents(local_client, "test", timestamp=5000)
        utils.set_s3_contents(s3_client, "hello", timestamp=6000)
        utils.set_s3_contents(
            s3_client, "baz", timestamp=8000, data="just syncing some stuff"
        )

        worker.sync()

        assert_existence(clients, ["foo"], False)
        assert_local_keys(clients, ["test", "bar", "baz", "hello"])
        assert_remote_timestamp(clients, "bar", 2000)
        assert_remote_timestamp(clients, "foo", 1000)
        assert_remote_timestamp(clients, "test", 5000)
        assert_remote_timestamp(clients, "hello", 6000)
        assert_remote_timestamp(clients, "baz", 8000)
        assert_contents(clients, "baz", b"just syncing some stuff")

        worker.sync()

    def test_three_way_sync(self, local_client, s3_client, local_client_2):
        utils.set_local_contents(local_client, "foo", timestamp=1000)
        utils.set_s3_contents(s3_client, "bar", timestamp=2000, data="red")
        utils.set_local_contents(local_client_2, "baz", timestamp=3000)

        worker_1 = sync.SyncWorker(local_client, s3_client)
        worker_2 = sync.SyncWorker(local_client_2, s3_client)
        worker_3 = sync.SyncWorker(local_client, local_client_2)

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        clients = [local_client, s3_client, local_client_2]

        assert_local_keys(clients, ["foo", "bar", "baz"])
        assert_contents(clients, "bar", b"red")
        assert_remote_timestamp(clients, "foo", 1000)
        assert_remote_timestamp(clients, "bar", 2000)
        assert_remote_timestamp(clients, "baz", 3000)

        utils.set_s3_contents(s3_client, "bar", timestamp=8000, data="green")

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        assert_local_keys(clients, ["foo", "bar", "baz"])
        assert_contents(clients, "bar", b"green")
        assert_remote_timestamp(clients, "foo", 1000)
        assert_remote_timestamp(clients, "bar", 8000)
        assert_remote_timestamp(clients, "baz", 3000)

        utils.delete_local(local_client_2, "foo")

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

        assert_existence(clients, ["foo"], False)
        assert_local_keys(clients, ["bar", "baz"])
        assert_remote_timestamp(clients, "foo", 1000)

        worker_1.sync()
        worker_2.sync()
        worker_3.sync()

    def test_ignore_conflicts(self, local_client, s3_client):
        utils.set_local_contents(local_client, "foo", timestamp=2000)
        utils.set_s3_contents(s3_client, "foo", timestamp=3000)
        utils.set_s3_contents(s3_client, "bar", timestamp=5600, data="usador")

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice="ignore")

        clients = [local_client, s3_client]
        assert_local_keys(clients, ["foo", "bar"])
        assert_remote_timestamp(clients, "foo", None)
        assert_remote_timestamp(clients, "bar", 5600)
        assert_existence(clients, ["foo", "bar"], True)
        assert_contents(clients, "bar", b"usador")

        # Check that none of the files have been modified due to ignore flag
        assert local_client.get_real_local_timestamp("foo") == 2000
        assert s3_client.get_real_local_timestamp("foo") == 3000

    def test_conflict_choose_first_client(self, local_client, s3_client):
        utils.set_local_contents(local_client, "foo", timestamp=2000, data="abc")
        utils.set_s3_contents(s3_client, "foo", timestamp=3000)

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice="1")

        clients = [local_client, s3_client]
        assert_local_keys(clients, ["foo"])
        # Chooses first client
        assert_remote_timestamp(clients, "foo", 2000)
        assert_contents(clients, "foo", b"abc")

    def test_conflict_choose_second_client(self, local_client, s3_client):
        utils.set_local_contents(local_client, "foo", timestamp=2000)
        utils.set_s3_contents(s3_client, "foo", timestamp=3000, data="123")

        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync(conflict_choice="2")

        clients = [local_client, s3_client]
        assert_local_keys(clients, ["foo"])
        # Chooses second client
        assert_remote_timestamp(clients, "foo", 3000)
        assert_contents(clients, "foo", b"123")

    def test_previously_deleted_now_created(self, local_client, s3_client):
        utils.set_s3_index(
            s3_client, {"foo": {"local_timestamp": None, "remote_timestamp": 4000}}
        )
        utils.set_local_index(
            local_client, {"foo": {"local_timstamp": None, "remote_timestamp": 4000}}
        )
        utils.set_local_contents(local_client, "foo", timestamp=7000)

        # Will create previously deleted file
        worker = sync.SyncWorker(local_client, s3_client)
        worker.sync()

        clients = [local_client, s3_client]
        assert_local_keys(clients, ["foo"])
        assert_remote_timestamp(clients, "foo", 7000)

        # delete the file again and check that it is successful
        utils.delete_local(local_client, "foo")

        worker.sync()
        assert_local_keys(clients, [])
        assert_remote_timestamp(clients, "foo", 7000)


class TestRunResolutions(object):
    def test_empty(self, local_client, s3_client):
        worker = sync.SyncWorker(local_client, s3_client)
        assert worker.run_resolutions({}) == []

    def test_invalid_action(self, local_client, s3_client):
        worker = sync.SyncWorker(local_client, s3_client)
        with pytest.raises(ValueError):
            worker.run_resolutions(
                {"foo": Resolution("UNKNOWN", None, None, None, None)}
            )

    def test_correct_output(self, local_client, s3_client):
        clients = [local_client, s3_client]
        worker = sync.SyncWorker(local_client, s3_client)

        utils.set_local_contents(local_client, "foo")
        utils.set_s3_contents(s3_client, "baz", timestamp=2000, data="testing")
        success = worker.run_resolutions(
            {
                "foo": Resolution(Resolution.DELETE, local_client, None, "foo", 1000),
                "baz": Resolution(
                    Resolution.CREATE, local_client, s3_client, "baz", 20
                ),
            }
        )
        assert sorted(success) == sorted(["foo", "baz"])
        assert_local_keys(clients, ["baz"])


class TestMoveClient(object):
    def test_correct_behaviour(self, local_client, s3_client):
        utils.set_s3_contents(s3_client, "art.txt", data="swirly abstract objects")

        worker = sync.SyncWorker(local_client, s3_client)

        resolution = Resolution(
            action=Resolution.CREATE,
            to_client=local_client,
            from_client=s3_client,
            key="art.txt",
            timestamp=6000,
        )
        worker.move_client(resolution)

        assert (
            utils.get_local_contents(local_client, "art.txt")
            == b"swirly abstract objects"
        )
        assert local_client.get_remote_timestamp("art.txt") == 6000



# -*- encoding: utf-8 -*-

from s4.commands.targets_command import TargetsCommand

from tests.utils import create_logger


class TestTargetsCommand(object):
    def test_empty(self, capsys):
        command = TargetsCommand(None, {"targets": {}}, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == err == ""

    def test_correct_output(self, capsys):
        config = {
            "targets": {
                "Personal": {
                    "s3_uri": "s3://mybackup/Personal",
                    "local_folder": "/home/user/Documents",
                },
                "Studies": {
                    "s3_uri": "s3://something/something/Studies",
                    "local_folder": "/media/backup/Studies",
                },
            }
        }

        command = TargetsCommand(None, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert err == ""
        assert out == (
            "Personal: [/home/user/Documents <=> s3://mybackup/Personal]\n"
            "Studies: [/media/backup/Studies <=> s3://something/something/Studies]\n"
        )



# -*- coding: utf-8 -*-
import datetime
import gzip
import json
import logging
import os
import zlib

import freezegun
import pytz

from s4.utils import to_timestamp


class InterruptedBytesIO(object):
    """
    Test helper class that imitates a BytesIO stream. Will return a stream of 0s for
    every requested read. When the number of times read was read reaches the parameter
    value of `interrupt_at`, a ValueError will be raised to simulate an interrupted IO
    transfer that could have occured
    """

    def __init__(self, interrupt_at=1):
        self.index = 0
        self.interrupt_at = interrupt_at

    def read(self, size):
        if self.index == self.interrupt_at:
            raise ValueError("Interrupted IO")
        else:
            self.index += 1
            return b"0" * size


def write_local(path, data=""):
    parent = os.path.dirname(path)
    if not os.path.exists(parent):
        os.makedirs(parent)
    with open(path, "w") as fp:
        fp.write(data)


def get_timestamp(year, month, day, hour, minute):
    return to_timestamp(
        datetime.datetime(year, month, day, hour, minute, tzinfo=pytz.UTC)
    )


def get_local_contents(client, key):
    with open(os.path.join(client.path, key), "rb") as fp:
        return fp.read()


def set_local_contents(client, key, timestamp=None, data=""):
    path = os.path.join(client.path, key)
    write_local(path, data)
    if timestamp is not None:
        os.utime(path, (timestamp, timestamp))


def delete_local(client, key):
    os.remove(os.path.join(client.path, key))


def set_local_index(client, data):
    with open(client.index_path(), "w") as fp:
        json.dump(data, fp)
    client.reload_index()


def write_s3(boto, bucket, key, data=""):
    boto.put_object(Bucket=bucket, Key=key, Body=data)


def set_s3_contents(s3_client, key, timestamp=None, data=""):
    if timestamp is None:
        freeze_time = datetime.datetime.utcnow()
    else:
        freeze_time = datetime.datetime.utcfromtimestamp(timestamp)

    with freezegun.freeze_time(freeze_time):
        write_s3(
            s3_client.boto, s3_client.bucket, os.path.join(s3_client.prefix, key), data
        )


def set_s3_index(s3_client, data, compression=None):
    body = json.dumps(data).encode("utf8")

    if compression == "gzip":
        body = gzip.compress(body)
    elif compression == "zlib":
        body = zlib.compress(body)

    s3_client.boto.put_object(
        Bucket=s3_client.bucket, Key=os.path.join(s3_client.prefix, ".index"), Body=body
    )
    s3_client.reload_index()


def create_logger():
    result = logging.getLogger("create_logger")
    result.setLevel(logging.INFO)
    result.handlers = []
    result.addHandler(logging.StreamHandler())
    return result



# -*- coding: utf-8 -*-
import datetime
import io
import os

import boto3
import freezegun
import mock
import pytest
from botocore.exceptions import ClientError
from moto import mock_s3

from s4.clients import SyncObject, s3
from s4.utils import to_timestamp

from tests import utils


class TestParseS3URI(object):
    def test_empty(self):
        assert s3.parse_s3_uri("") is None

    def test_incorrect_format(self):
        assert s3.parse_s3_uri("nos3infront/some/path") is None
        assert s3.parse_s3_uri(":/232red+-32") is None

    def test_no_path(self):
        assert s3.parse_s3_uri("s3://bucket") is None

    def test_correct_output(self):
        actual_output = s3.parse_s3_uri("s3://fruit.bowl/apples/and/oranges/")
        assert actual_output.bucket == "fruit.bowl"
        assert actual_output.key == "apples/and/oranges/"


class TestIsIgnoredFile(object):
    def test_empty(self):
        assert s3.is_ignored_key("foo/bar/baz", []) is False

    def test_subdirectory(self):
        assert s3.is_ignored_key("foo/.git/baz", [".git"]) is True

    def test_file(self):
        assert s3.is_ignored_key("foo/ignoreme", ["ignore*"]) is True


class TestS3SyncClient(object):
    def test_get_client_name(self, s3_client):
        assert s3_client.get_client_name() == "s3"

    def test_repr(self):
        boto_client = boto3.client(
            "s3", aws_access_key_id="", aws_secret_access_key="", aws_session_token=""
        )

        client = s3.S3SyncClient(boto_client, "testbucket", "foo/bar")
        assert repr(client) == "S3SyncClient<testbucket, foo/bar>"

    @mock_s3
    def test_get_uri(self):
        s3_client = boto3.client(
            "s3", aws_access_key_id="", aws_secret_access_key="", aws_session_token=""
        )

        client = s3.S3SyncClient(s3_client, "testbucket", "foo/bar")
        assert client.get_uri() == "s3://testbucket/foo/bar/"
        assert client.get_uri("apples.txt") == "s3://testbucket/foo/bar/apples.txt"

    @mock_s3
    def test_index_path(self):
        s3_client = boto3.client(
            "s3", aws_access_key_id="", aws_secret_access_key="", aws_session_token=""
        )

        client = s3.S3SyncClient(s3_client, "testbucket", "foo/bar")
        assert client.index_path() == "foo/bar/.index"

    def test_put(self, s3_client):
        # given
        data = b"munchkin"

        # when
        input_object = SyncObject(io.BytesIO(data), len(data), 4000)
        s3_client.put("something/boardgame.rst", input_object)

        # then
        resp = s3_client.boto.get_object(
            Bucket=s3_client.bucket,
            Key=os.path.join(s3_client.prefix, "something/boardgame.rst"),
        )
        assert resp["Body"].read() == data
        assert s3_client.get_remote_timestamp("something/boardgame.rst") == 4000

    def test_put_callback(self, s3_client):
        # given
        data = b"eygon of carim"
        mock_callback = mock.MagicMock()

        # when
        input_object = SyncObject(io.BytesIO(data), len(data), 4000)
        s3_client.put("dark/souls.rst", input_object, callback=mock_callback)

        assert mock_callback.call_count == 1
        mock_callback.assert_called_with(len(data))

    def test_get(self, s3_client):
        # given
        data = b"#000000"
        frozen_time = datetime.datetime(
            2016, 10, 23, 10, 30, tzinfo=datetime.timezone.utc
        )
        with freezegun.freeze_time(frozen_time):
            s3_client.boto.put_object(
                Bucket=s3_client.bucket,
                Key=os.path.join(s3_client.prefix, "black.color"),
                Body=data,
            )

        # when
        output_object = s3_client.get("black.color")

        # then
        assert output_object.fp.read() == data
        assert output_object.total_size == len(data)
        assert output_object.timestamp == to_timestamp(frozen_time)

    def test_get_non_existent(self, s3_client):
        assert s3_client.get("idontexist.md") is None

    def test_put_get(self, s3_client):
        # given
        data = b"woooooooooshhh"
        input_object = SyncObject(io.BytesIO(data), len(data), 4000)
        frozen_time = datetime.datetime(
            2016, 10, 23, 10, 30, tzinfo=datetime.timezone.utc
        )

        # when
        with freezegun.freeze_time(frozen_time):
            s3_client.put("something/woosh.gif", input_object)

        # then
        output_object = s3_client.get("something/woosh.gif")
        assert output_object.fp.read() == data
        assert output_object.timestamp == to_timestamp(frozen_time)

    def test_delete(self, s3_client):
        # given
        s3_client.boto.put_object(
            Bucket=s3_client.bucket,
            Key=os.path.join(s3_client.prefix, "war.png"),
            Body="bang",
        )

        # when
        assert s3_client.delete("war.png") is True

        # then
        assert s3_client.get("war.png") is None
        with pytest.raises(ClientError) as exc:
            s3_client.boto.head_object(
                Bucket=s3_client.bucket, Key=os.path.join(s3_client.prefix, "war.png")
            )
        assert exc.value.response["Error"]["Code"] == "404"

    def test_delete_non_existent(self, s3_client):
        assert s3_client.delete("idontexist.png") is False

    def test_get_local_keys(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, "war.png")
        utils.set_s3_contents(s3_client, "this/is/fine")
        utils.set_s3_contents(s3_client, ".index")
        utils.write_s3(
            s3_client.boto,
            s3_client.bucket,
            os.path.join("iamsomethingelse", "ishouldnotshow.txt"),
        )

        # when
        actual_output = s3_client.get_local_keys()

        # then
        expected_output = ["war.png", "this/is/fine"]
        assert sorted(actual_output) == sorted(expected_output)

    def test_get_index_keys(self, s3_client):
        utils.set_s3_index(
            s3_client,
            {
                "cow": {"local_timestamp": 4000, "remote_timestamp": 3000},
                "chicken": {"local_timestamp": 4000, "remote_timestamp": 3000},
            },
        )
        assert sorted(s3_client.get_index_keys()) == sorted(["cow", "chicken"])

    @pytest.mark.parametrize(
        ("magic_result", "compression"),
        [
            ("text/plain", None),
            ("application/json", None),
            ("application/gzip", "gzip"),
            ("application/x-gzip", "gzip"),
            ("application/zlib", "zlib"),
            ("application/octet-stream", "gzip"),
            ("application/octet-stream", "zlib"),
        ],
    )
    def test_get_index_timestamps(self, s3_client, compression, magic_result):
        with mock.patch("magic.from_buffer") as from_buffer:
            # given
            from_buffer.return_value = magic_result

            utils.set_s3_index(
                s3_client,
                {
                    "hello": {"remote_timestamp": 1234, "local_timestamp": 1200},
                    "world": {"remote_timestamp": 5000},
                },
                compression=compression,
            )

            # then
            assert s3_client.get_remote_timestamp("hello") == 1234
            assert s3_client.get_index_local_timestamp("hello") == 1200

            assert s3_client.get_remote_timestamp("world") == 5000
            assert s3_client.get_index_local_timestamp("world") is None

    def test_get_all_index_timestamps(self, s3_client):
        # given
        utils.set_s3_index(
            s3_client,
            {"hello": {"local_timestamp": 1200}, "world": {"local_timestamp": 4000}},
        )

        # then
        expected_output = {"hello": 1200, "world": 4000}
        actual_output = s3_client.get_all_index_local_timestamps()
        assert actual_output == expected_output

    def test_update_index(self, s3_client):
        # given
        utils.set_s3_index(
            s3_client,
            {
                "red": {"remote_timestamp": 1234, "local_timestamp": 1200},
                "green": {"remote_timestamp": 5000},
            },
        )

        utils.set_s3_contents(s3_client, "red", timestamp=5001)
        utils.set_s3_contents(s3_client, "yellow", timestamp=1000)
        utils.set_s3_contents(s3_client, "orange", timestamp=2000)

        # when
        s3_client.update_index()

        # then
        expected_index = {
            "red": {"remote_timestamp": 1234, "local_timestamp": 5001},
            "green": {"remote_timestamp": 5000, "local_timestamp": None},
            "yellow": {"remote_timestamp": None, "local_timestamp": 1000},
            "orange": {"remote_timestamp": None, "local_timestamp": 2000},
        }
        assert s3_client.index == expected_index

    def test_get_real_local_timestamp(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, "orange", timestamp=2000)

        # then
        assert s3_client.get_real_local_timestamp("orange") == 2000
        assert s3_client.get_real_local_timestamp("idontexist") is None

    def test_get_all_real_local_timestamps(self, s3_client):
        # given
        utils.set_s3_contents(s3_client, "chocolate/oranges", timestamp=2600)
        utils.set_s3_contents(s3_client, "gummy/bears", timestamp=2400)
        utils.set_s3_contents(s3_client, "carrot_cake", timestamp=2000)
        utils.set_s3_contents(s3_client, ".index", timestamp=2043)
        s3_client.reload_index()

        # then
        expected_output = {
            "chocolate/oranges": 2600,
            "gummy/bears": 2400,
            "carrot_cake": 2000,
        }
        actual_output = s3_client.get_all_real_local_timestamps()
        assert actual_output == expected_output

    def test_set_index_timestamps(self, s3_client):
        # given
        utils.set_s3_index(
            s3_client, {"red": {"remote_timestamp": 1234, "local_timestamp": 1200}}
        )

        # when
        s3_client.set_index_local_timestamp("red", 3000)
        s3_client.set_remote_timestamp("red", 4000)
        s3_client.set_index_local_timestamp("green", 5000)
        s3_client.set_remote_timestamp("yellow", 6000)

        # then
        expected_index = {
            "red": {"local_timestamp": 3000, "remote_timestamp": 4000},
            "green": {"local_timestamp": 5000},
            "yellow": {"remote_timestamp": 6000},
        }
        assert s3_client.index == expected_index

    def test_interrupted_put(self, s3_client):
        utils.set_s3_contents(s3_client, "keychain", data="iamsomedatathatexists")

        sync_object = SyncObject(utils.InterruptedBytesIO(), 900000, 3000)

        with pytest.raises(ValueError):
            s3_client.put("keychain", sync_object)

        result = s3_client.get("keychain")
        assert result.fp.read() == b"iamsomedatathatexists"

    def test_ignore_files(self, s3_client):
        utils.set_s3_contents(
            s3_client,
            ".syncignore",
            timestamp=4000,
            data=("*~\n" "trashdirectory\n" "*py[oc]\n" ".git\n"),
        )
        s3_client.reload_ignore_files()

        utils.set_s3_contents(s3_client, ".zshrc~")
        utils.set_s3_contents(s3_client, "trashdirectory")
        utils.set_s3_contents(s3_client, "foo/mobile.py", timestamp=1290)
        utils.set_s3_contents(s3_client, "foo/mobile.pyc")
        utils.set_s3_contents(s3_client, "foo/mobile.pyo")
        utils.set_s3_contents(
            s3_client, "foo/.git/7f04e3b69963e5d51b479b30591ab1c2b6a3e7"
        )
        utils.set_s3_contents(s3_client, ".zshrc", timestamp=9999)

        s3_client.update_index()

        expected_index = {
            ".syncignore": {"local_timestamp": 4000, "remote_timestamp": None},
            ".zshrc": {"local_timestamp": 9999, "remote_timestamp": None},
            "foo/mobile.py": {"local_timestamp": 1290, "remote_timestamp": None},
        }
        assert s3_client.index == expected_index



# -*- encoding: utf-8 -*-
import argparse
import json

import mock

from s4.commands.edit_command import EditCommand

from tests import utils


@mock.patch("s4.utils.get_input")
class TestEditCommand(object):
    def test_no_targets(self, get_input, capsys):
        command = EditCommand(None, {}, utils.create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            "You have not added any targets yet\n" 'Use the "add" command to do this\n'
        )

    def test_missing_target(self, get_input, capsys):
        args = argparse.Namespace(target="idontexist")
        config = {"targets": {"foo": {}}}
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ('"idontexist" is an unknown target\n' "Choices are: ['foo']\n")

    def test_no_changes(self, get_input, config_file):
        fake_stream = ["", "", "", "", "", ""]
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target="foo")
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mark/documents",
                    "endpoint_url": "https://customurl.com",
                    "s3_uri": "s3://buckets/mybackup",
                    "aws_access_key_id": "23123123123313",
                    "aws_secret_access_key": "edwdadwadwadwdd",
                    "region_name": "eu-west-1",
                }
            }
        }
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        with open(config_file, "r") as fp:
            config = json.load(fp)

        expected_config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mark/documents",
                    "endpoint_url": "https://customurl.com",
                    "s3_uri": "s3://buckets/mybackup",
                    "aws_access_key_id": "23123123123313",
                    "aws_secret_access_key": "edwdadwadwadwdd",
                    "region_name": "eu-west-1",
                }
            }
        }
        assert expected_config == config

    def test_correct_output(self, get_input, config_file):
        fake_stream = [
            "/home/user/Documents",
            "https://example.com",
            "s3://buckets/mybackup222",
            "9999999999",
            "bbbbbbbbbbbbbbbbbbbbbbbb",
            "eu-west-2",
        ]
        get_input.side_effect = fake_stream

        args = argparse.Namespace(target="foo")
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mark/documents",
                    "endpoint_url": "",
                    "s3_uri": "s3://buckets/mybackup",
                    "aws_access_key_id": "23123123123313",
                    "aws_secret_access_key": "edwdadwadwadwdd",
                    "region_name": "eu-west-1",
                }
            }
        }
        command = EditCommand(args, config, utils.create_logger())
        command.run()

        with open(config_file, "r") as fp:
            config = json.load(fp)

        expected_config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/user/Documents",
                    "endpoint_url": "https://example.com",
                    "s3_uri": "s3://buckets/mybackup222",
                    "aws_access_key_id": "9999999999",
                    "aws_secret_access_key": "bbbbbbbbbbbbbbbbbbbbbbbb",
                    "region_name": "eu-west-2",
                }
            }
        }
        assert expected_config == config



# -*- coding: utf-8 -*-

import os
import shutil
import tempfile

import boto3
import mock
import moto
import pytest
from faker import Faker

from s4.clients.local import LocalSyncClient
from s4.clients.s3 import S3SyncClient

fake = Faker()


@pytest.fixture
def config_file():
    fd, temp_path = tempfile.mkstemp()
    mocker = mock.patch("s4.utils.CONFIG_FILE_PATH", temp_path)
    mocker.start()
    yield temp_path
    mocker.stop()
    os.close(fd)
    os.remove(temp_path)


@pytest.fixture
def local_client():
    folder = tempfile.mkdtemp()
    yield LocalSyncClient(folder)
    shutil.rmtree(folder)


@pytest.fixture
def local_client_2():
    folder = tempfile.mkdtemp()
    yield LocalSyncClient(folder)
    shutil.rmtree(folder)


@pytest.fixture
def s3_client():
    mock = moto.mock_s3()
    mock.start()
    boto_client = boto3.client(
        "s3", aws_access_key_id="", aws_secret_access_key="", aws_session_token=""
    )
    bucket_name = fake.user_name()
    boto_client.create_bucket(Bucket=bucket_name)
    yield S3SyncClient(boto_client, bucket_name, fake.uri_path())
    mock.stop()



# -*- encoding: utf-8 -*-

import argparse
import json
import os

import mock

from s4.commands.add_command import AddCommand

from tests import utils


@mock.patch("s4.utils.get_input")
class TestAddCommand(object):
    def test_correct_behaviour(self, get_input, config_file):
        get_input.side_effect = [
            "/home/user/Documents",
            None,
            "mybucket",
            "Documents",
            "eu-west-2",
            "aaaaaaaaaaaaaaaaaaaaaaaa",
            "bbbbbbbbbbbbbbbbbbbbbbbb",
            "",
        ]
        args = argparse.Namespace(copy_target_credentials=None)

        command = AddCommand(args, {"targets": {}}, utils.create_logger())
        command.run()

        with open(config_file, "r") as fp:
            new_config = json.load(fp)

        expected_config = {
            "targets": {
                "Documents": {
                    "local_folder": "/home/user/Documents",
                    "endpoint_url": None,
                    "s3_uri": "s3://mybucket/Documents",
                    "aws_access_key_id": "aaaaaaaaaaaaaaaaaaaaaaaa",
                    "aws_secret_access_key": "bbbbbbbbbbbbbbbbbbbbbbbb",
                    "region_name": "eu-west-2",
                }
            }
        }
        assert new_config == expected_config

    def test_default_local_folder(self, get_input, config_file):
        get_input.side_effect = [
            None,
            None,
            "mybucket",
            "Documents",
            "eu-west-2",
            "aaaaaaaaaaaaaaaaaaaaaaaa",
            "bbbbbbbbbbbbbbbbbbbbbbbb",
            "",
        ]
        args = argparse.Namespace(copy_target_credentials=None)

        command = AddCommand(args, {"targets": {}}, utils.create_logger())
        command.run()

        with open(config_file, "r") as fp:
            config = json.load(fp)

        assert config["targets"]["Documents"]["local_folder"] == os.getcwd()

    def test_copy_target_credentials(self, get_input, config_file):
        get_input.side_effect = [
            "/home/user/Animals",
            None,
            "mybucket",
            "Zoo",
            "us-west-2",
            "Beasts",
        ]
        args = argparse.Namespace(copy_target_credentials="bar")

        command = AddCommand(
            args,
            {
                "targets": {
                    "bar": {
                        "aws_secret_access_key": "bar-secretz",
                        "aws_access_key_id": "so-much-bar",
                    }
                }
            },
            utils.create_logger(),
        )
        command.run()

        with open(config_file, "r") as fp:
            new_config = json.load(fp)

        expected_config = {
            "targets": {
                "bar": {
                    "aws_access_key_id": "so-much-bar",
                    "aws_secret_access_key": "bar-secretz",
                },
                "Beasts": {
                    "local_folder": "/home/user/Animals",
                    "endpoint_url": None,
                    "s3_uri": "s3://mybucket/Zoo",
                    "aws_access_key_id": "so-much-bar",
                    "aws_secret_access_key": "bar-secretz",
                    "region_name": "us-west-2",
                },
            }
        }
        assert new_config == expected_config

    def test_copy_target_credentials_bad_target(self, get_input, capsys):
        get_input.side_effect = [
            "/home/user/Animals",
            "",
            "mybucket",
            "Zoo",
            "us-west-2",
            "Beasts",
        ]
        args = argparse.Namespace(copy_target_credentials="Foo")

        command = AddCommand(args, {"targets": {"bar": {}}}, utils.create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ('"Foo" is an unknown target\n' "Choices are: ['bar']\n")

    def test_custom_target_name(self, get_input, config_file):
        get_input.side_effect = [
            "/home/user/Music",
            None,
            "mybucket",
            "Musiccccc",
            "us-west-1",
            "1234567890",
            "abcdefghij",
            "Tunes",
        ]
        args = argparse.Namespace(copy_target_credentials=None)

        command = AddCommand(args, {"targets": {}}, utils.create_logger())
        command.run()

        with open(config_file, "r") as fp:
            new_config = json.load(fp)

        expected_config = {
            "targets": {
                "Tunes": {
                    "local_folder": "/home/user/Music",
                    "endpoint_url": None,
                    "s3_uri": "s3://mybucket/Musiccccc",
                    "aws_access_key_id": "1234567890",
                    "aws_secret_access_key": "abcdefghij",
                    "region_name": "us-west-1",
                }
            }
        }
        assert new_config == expected_config



# -*- encoding: utf-8 -*-

import argparse

import mock
import pytest
from inotify_simple import Event, flags

from s4.commands.daemon_command import DaemonCommand

from tests.utils import create_logger


class FakeINotify(object):
    def __init__(self, events, wd_map):
        self.events = events
        self.wd_map = wd_map

    def add_watches(self, *args, **kwargs):
        return self.wd_map

    def read(self, *args, **kwargs):
        return self.events


@mock.patch("s4.sync.SyncWorker")
@mock.patch("s4.commands.daemon_command.INotifyRecursive")
class TestDaemonCommand(object):
    def single_term(self, index):
        """Simple terminator for the daemon command"""
        return index >= 1

    @mock.patch("s4.commands.daemon_command.supported", False)
    def test_os_not_supported(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts="ignore", read_delay=0)

        command = DaemonCommand(args, {}, create_logger())
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            "Cannot run INotify on your operating system\n"
            "Only Linux machines are officially supported for this command\n"
        )
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_no_targets(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts="ignore", read_delay=0)

        command = DaemonCommand(args, {"targets": {}}, create_logger())
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ("No targets available\n" 'Use "add" command first\n')
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_wrong_target(self, INotifyRecursive, SyncWorker, capsys):
        args = argparse.Namespace(targets=["foo"], conflicts="ignore", read_delay=0)

        command = DaemonCommand(args, {"targets": {"bar": {}}}, create_logger())
        command.run(terminator=self.single_term)

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ("Unknown target: foo\n")
        assert SyncWorker.call_count == 0
        assert INotifyRecursive.call_count == 0

    @pytest.mark.timeout(5)
    def test_specific_target(self, INotifyRecursive, SyncWorker):
        INotifyRecursive.return_value = FakeINotify(
            events={
                Event(wd=1, mask=flags.CREATE, cookie=None, name="hello.txt"),
                Event(wd=2, mask=flags.CREATE, cookie=None, name="bar.txt"),
            },
            wd_map={1: "/home/jon/code/", 2: "/home/jon/code/hoot"},
        )

        args = argparse.Namespace(targets=["foo"], conflicts="ignore", read_delay=0)
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/jon/code",
                    "s3_uri": "s3://bucket/code",
                    "aws_secret_access_key": "23232323",
                    "aws_access_key_id": "########",
                    "region_name": "eu-west-2",
                },
                "bar": {},
            }
        }
        command = DaemonCommand(args, config, create_logger())
        command.run(terminator=self.single_term)

        assert SyncWorker.call_count == 2
        assert INotifyRecursive.call_count == 1



# -*- coding: utf-8 -*-

import gzip
import json
import zlib

import mock
import pytest

from s4 import utils


class TestTryDecompress:
    def test_bad_value(self):
        with pytest.raises(ValueError) as exc:
            utils.try_decompress(b"Not compressed data")
        assert str(exc.value) == "Unknown compression format"

    def test_gzip(self):
        body = gzip.compress(b"some data")
        assert utils.try_decompress(body) == b"some data"

    def test_zlib(self):
        body = zlib.compress(b"some data")
        assert utils.try_decompress(body) == b"some data"


@mock.patch("getpass.getpass")
@mock.patch("builtins.input")
class TestGetInput:
    def test_required(self, input_fn, getpass):
        input_fn.side_effect = ["", "", "something"]

        result = utils.get_input("give me some info", required=True)

        assert result == "something"
        assert input_fn.call_count == 3
        assert getpass.call_count == 0

    def test_not_secret(self, input_fn, getpass):
        input_fn.return_value = "foo"

        result = utils.get_input("give me some info", secret=False)

        assert result == "foo"

        assert getpass.call_count == 0
        assert input_fn.call_count == 1

    def test_blank(self, input_fn, getpass):
        input_fn.return_value = ""

        result = utils.get_input("give me some info", blank=True)

        assert result is None

        assert getpass.call_count == 0
        assert input_fn.call_count == 1

    def test_secret(self, input_fn, getpass):
        getpass.return_value = "bar"

        result = utils.get_input("give me some secret info", secret=True)

        assert result == "bar"

        assert getpass.call_count == 1
        assert input_fn.call_count == 0


class TestGetConfigFile(object):
    @mock.patch("s4.utils.CONFIG_FILE_PATH", "/i/dont/exist")
    def test_no_file(self):
        assert utils.get_config() == {"targets": {}}

    def test_correct_output(self, config_file):
        with open(config_file, "w") as fp:
            json.dump({"local_folder": "/home/someone/something"}, fp)

        assert utils.get_config() == {"local_folder": "/home/someone/something"}



#! -*- encoding: utf8 -*-

import os
import shutil
import subprocess
import tempfile


def show_diff(client_1, client_2, key):
    if shutil.which("diff") is None:
        print('Missing required "diff" executable.')
        print("Install this using your distribution's package manager")
        return

    if shutil.which("less") is None:
        print('Missing required "less" executable.')
        print("Install this using your distribution's package manager")
        return

    so_1 = client_1.get(key)
    data_1 = so_1.fp.read()
    so_1.fp.close()

    so_2 = client_2.get(key)
    data_2 = so_2.fp.read()
    so_2.fp.close()

    fd_1, path_1 = tempfile.mkstemp()
    fd_2, path_2 = tempfile.mkstemp()
    fd_3, path_3 = tempfile.mkstemp()

    with open(path_1, "wb") as fp:
        fp.write(data_1)
    with open(path_2, "wb") as fp:
        fp.write(data_2)

    # This is a lot faster than the difflib found in python
    with open(path_3, "wb") as fp:
        subprocess.call(
            [
                "diff",
                "-u",
                "--label",
                client_1.get_uri(key),
                path_1,
                "--label",
                client_2.get_uri(key),
                path_2,
            ],
            stdout=fp,
        )

    subprocess.call(["less", path_3])

    os.close(fd_1)
    os.close(fd_2)
    os.close(fd_3)

    os.remove(path_1)
    os.remove(path_2)
    os.remove(path_3)



# -*- encoding: utf-8 -*-

import argparse

from s4.commands.ls_command import LsCommand

from tests.utils import create_logger, get_timestamp, set_local_index, set_s3_index


class TestLsCommand(object):
    def test_empty_config(self, capsys):
        args = argparse.Namespace(target="idontexist")
        command = LsCommand(args, {}, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            "You have not added any targets yet\n" 'Use the "add" command to do this\n'
        )

    def test_missing_target(self, capsys):
        args = argparse.Namespace(target="idontexist")
        config = {"targets": {"foo": {}, "bar": {}}}
        command = LsCommand(args, config, create_logger())
        command.run()
        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            '"idontexist" is an unknown target\n' "Choices are: ['bar', 'foo']\n"
        )

    def test_correct_output_empty(self, s3_client, local_client, capsys):
        config = {
            "targets": {
                "foo": {
                    "local_folder": local_client.get_uri(),
                    "s3_uri": s3_client.get_uri(),
                    "aws_access_key_id": "",
                    "aws_secret_access_key": "",
                    "region_name": "eu-west-2",
                }
            }
        }
        args = argparse.Namespace(target="foo", sort_by="key", descending=False)
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            "key    local    s3\n" "-----  -------  ----\n" "Total Size: 0.00Mb\n"
        )

    def test_correct_output_nonempty(self, s3_client, local_client, capsys):
        config = {
            "targets": {
                "foo": {
                    "local_folder": local_client.get_uri(),
                    "s3_uri": s3_client.get_uri(),
                    "aws_access_key_id": "",
                    "aws_secret_access_key": "",
                    "region_name": "eu-west-2",
                }
            }
        }
        set_s3_index(
            s3_client,
            {
                "milk": {"local_timestamp": get_timestamp(1989, 10, 23, 11, 30)},
                "honey": {"local_timestamp": get_timestamp(2016, 12, 12, 8, 30)},
                "ginger": {"local_timestamp": None},
            },
        )
        set_local_index(
            local_client,
            {
                "milk": {"local_timestamp": get_timestamp(2016, 12, 12, 8, 30)},
                "honey": {"local_timestamp": get_timestamp(2016, 11, 10, 18, 40)},
                "lemon": {"local_timestamp": get_timestamp(2017, 2, 2, 8, 30)},
                "ginger": {"local_timestamp": None},
            },
        )

        args = argparse.Namespace(
            target="foo", sort_by="key", show_all=False, descending=False
        )
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            "key    local                s3\n"
            "-----  -------------------  -------------------\n"
            "honey  2016-11-10 18:40:00  2016-12-12 08:30:00\n"
            "lemon  2017-02-02 08:30:00\n"
            "milk   2016-12-12 08:30:00  1989-10-23 11:30:00\n"
            "Total Size: 0.00Mb\n"
        )

    def test_show_all(self, s3_client, local_client, capsys):
        config = {
            "targets": {
                "foo": {
                    "local_folder": local_client.get_uri(),
                    "s3_uri": s3_client.get_uri(),
                    "aws_access_key_id": "",
                    "aws_secret_access_key": "",
                    "region_name": "eu-west-2",
                }
            }
        }
        set_s3_index(
            s3_client,
            {
                "cheese": {"local_timestamp": get_timestamp(2017, 12, 12, 8, 30)},
                "crackers": {"local_timestamp": None},
            },
        )
        set_local_index(
            local_client,
            {
                "cheese": {"local_timestamp": get_timestamp(2017, 2, 2, 8, 30)},
                "crackers": {"local_timestamp": None},
            },
        )

        args = argparse.Namespace(
            target="foo", sort_by="key", show_all=True, descending=False
        )
        command = LsCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()

        assert err == ""
        assert out == (
            "key       local                s3\n"
            "--------  -------------------  -------------------\n"
            "cheese    2017-02-02 08:30:00  2017-12-12 08:30:00\n"
            "crackers  <deleted>\n"
            "Total Size: 0.00Mb\n"
        )



#! -*- encoding: utf8 -*
import os

from s4 import utils
from s4.commands import Command


class AddCommand(Command):
    def run(self):
        target = self.args.copy_target_credentials

        all_targets = list(self.config["targets"].keys())
        if target is not None and target not in all_targets:
            self.logger.info('"%s" is an unknown target', target)
            self.logger.info("Choices are: %s", all_targets)
            return

        if target is not None:
            target_config = self.config["targets"][target]

            aws_access_key_id = target_config["aws_access_key_id"]
            aws_secret_access_key = target_config["aws_secret_access_key"]
        else:
            aws_access_key_id = None
            aws_secret_access_key = None

        self.logger.info("To add a new target, please enter the following\n")

        if self.args.local_folder and self.args.endpoint_url and self.args.bucket and self.args.path and \
                self.args.access_key and self.args.secret_key:
            local_folder = self.args.local_folder
            endpoint_url = self.args.endpoint_url
            bucket = self.args.bucket
            path = self.args.path
            aws_access_key_id = self.args.access_key
            aws_secret_access_key = self.args.secret_key
            region_name = ""
            if self.args.target_name:
                name = self.args.target_name
            else:
                name = "default"
        else:
            local_folder = utils.get_input(
                "local folder path to sync [leave blank for current folder]: "
            )
            endpoint_url = utils.get_input(
                "endpoint url [leave blank for AWS]: ", blank=True
            )
            bucket = utils.get_input("S3 Bucket [required]: ", required=True)
            path = utils.get_input("S3 Path: ")
            region_name = utils.get_input(
                "region name [leave blank if unknown]: ", blank=True
            )

            if aws_access_key_id is None:
                aws_access_key_id = utils.get_input(
                    "AWS Access Key ID [required]: ", required=True
                )
                aws_secret_access_key = utils.get_input(
                    "AWS Secret Access Key [required]: ", secret=True, required=True
                )

            default_name = os.path.basename(path)
            name = utils.get_input(
                "Provide a name for this entry [leave blank to default to '{}']: ".format(
                    default_name
                )
            )
            name = name or default_name
            if not local_folder:
                local_folder = os.getcwd()

            local_folder = os.path.expanduser(local_folder)
            local_folder = os.path.abspath(local_folder)

        self.config["targets"][name] = {
            "local_folder": local_folder,
            "endpoint_url": endpoint_url,
            "s3_uri": "s3://{}/{}".format(bucket, path),
            "region_name": region_name,
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
        }
        utils.set_config(self.config)

        self.logger.info(
            "\nTarget has been added. Start syncing with the 'sync' command"
        )
        self.logger.info(
            "You can edit anything you have entered here using the 'edit' command"
        )



#! -*- encoding: utf-8 -*-
from collections import defaultdict

from s4.commands import Command
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer
import time

# Don't crash on import if the underlying operating system does not support INotify
try:
    from inotify_simple import flags
    from s4.inotify_recursive import INotifyRecursive

    supported = True
except (OSError, ModuleNotFoundError):
    supported = False


# class DaemonCommand(Command):
#     def run(self, terminator=lambda x: False):
#         if not supported:
#             self.logger.info("Cannot run INotify on your operating system")
#             self.logger.info(
#                 "Only Linux machines are officially supported for this command"
#             )
#             return
#
#         all_targets = list(self.config["targets"].keys())
#         if not self.args.targets:
#             targets = all_targets
#         else:
#             targets = self.args.targets
#
#         if not targets:
#             self.logger.info("No targets available")
#             self.logger.info('Use "add" command first')
#             return
#
#         for target in targets:
#             if target not in self.config["targets"]:
#                 self.logger.info("Unknown target: %s", target)
#                 return
#
#         notifier = INotifyRecursive()
#         watch_flags = flags.CREATE | flags.DELETE | flags.MODIFY
#
#         watch_map = {}
#
#         for target in targets:
#             entry = self.config["targets"][target]
#             path = entry["local_folder"]
#             self.logger.info("Watching %s", path)
#             for wd in notifier.add_watches(path.encode("utf8"), watch_flags):
#                 watch_map[wd] = target
#
#             # Check for any pending changes
#             worker = self.get_sync_worker(target)
#             worker.sync(conflict_choice=self.args.conflicts)
#
#         index = 0
#
#         while not terminator(index):
#             index += 1
#
#             to_run = defaultdict(set)
#
#             for event in notifier.read(read_delay=self.args.read_delay):
#                 target = watch_map[event.wd]
#
#                 # Don't bother running for .index
#                 if event.name not in (".index", ".s4lock"):
#                     to_run[target].add(event.name)
#
#             for target, keys in to_run.items():
#                 worker = self.get_sync_worker(target)
#
#                 # Should ideally be setting keys to sync
#                 self.logger.info("Syncing {}".format(worker))
#                 worker.sync(conflict_choice=self.args.conflicts)


# override by Minghuan Ma
class DaemonCommand(Command):

    """
    One-way continuous sync from localpath to s3 path (uses a file watcher called watchdog)
    """

    def run(self, terminator=lambda x: False):
        print("Daemon command running")

        all_targets = list(self.config["targets"].keys())
        if not self.args.targets:
            targets = all_targets
        else:
            targets = self.args.targets

        if not targets:
            self.logger.info("No targets available")
            self.logger.info('Use "add" command first')
            return

        for target in targets:
            if target not in self.config["targets"]:
                self.logger.info("Unknown target: %s", target)
                return

        for target in targets:
            entry = self.config["targets"][target]
            path = entry["local_folder"]

            # do sync at begainning
            worker = self.get_sync_worker(target)
            # conflict_choice参数："1"/"2"/"ignore"
            # conflict_choice="1": 如果某些文件s3dir有, localdir没有, 删除s3路径上的差别文件
            # conflict_choice="2": 如果某些文件s3dir有, localdir没有, 将差别文件下拉到localdir
            # conflict_choice="ignore": 忽略差别文件
            worker.sync(conflict_choice="ignore")

            event_handler = FileWatchHandler(worker, self.args)
            observer = Observer()
            observer.schedule(event_handler, path, recursive=True)
            observer.start()
            self.logger.info("Watching %s", path)

            try:
                while True:
                    time.sleep(2)
            except KeyboardInterrupt:
                observer.stop()
            observer.join()


class FileWatchHandler(PatternMatchingEventHandler):

    def __init__(self, sync_worker, args):
        super().__init__(ignore_patterns=[".index", ".s4lock"])
        self.sync_worker = sync_worker
        self.args = args

    def on_any_event(self, event):

        if event.is_directory:
            return

        print("Watchdog received [{}] event for [{}] - [{}]"
              .format(event.event_type, "directory" if event.is_directory else "file", event.src_path))
        self.sync_worker.sync(conflict_choice=self.args.conflicts)


# -*- coding: utf-8 -*-

import datetime
import getpass
import gzip
import json
import os
import zlib

CONFIG_FOLDER_PATH = os.path.expanduser("~/.config/s4")
CONFIG_FILE_PATH = os.path.join(CONFIG_FOLDER_PATH, "sync.conf")


def try_decompress(body):
    # Attempt multiple levels of fallback in case we cannot
    # figure out the compression type using `magic`
    try:
        return gzip.decompress(body)
    except OSError:
        pass

    try:
        return zlib.decompress(body)
    except zlib.error:
        raise ValueError("Unknown compression format")


def to_timestamp(dt):
    epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    return (dt - epoch) / datetime.timedelta(seconds=1)


def get_input(*args, secret=False, required=False, blank=False, **kwargs):
    """
    secret: Don't show user input when they are typing.
    required: Keep prompting if the user enters an empty value.
    blank: turn all empty strings into None.
    """

    while True:
        if secret:
            value = getpass.getpass(*args, **kwargs)
        else:
            value = input(*args, **kwargs)

        if blank:
            value = value if value else None

        if not required or value:
            break

    return value


def get_config():
    if not os.path.exists(CONFIG_FILE_PATH):
        return {"targets": {}}

    with open(CONFIG_FILE_PATH, "r") as fp:
        config = json.load(fp)
    return config


def set_config(config):
    if not os.path.exists(CONFIG_FOLDER_PATH):
        os.makedirs(CONFIG_FOLDER_PATH)

    with open(CONFIG_FILE_PATH, "w") as fp:
        json.dump(config, fp, indent=5)



# -*- encoding: utf-8 -*-

import argparse
import json

from s4.commands.rm_command import RmCommand

from tests.utils import create_logger


class TestRemoveCommand(object):
    def test_empty(self, capsys, config_file):
        args = argparse.Namespace(target="foo")
        command = RmCommand(args, {}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ("You have not added any targets yet\n")

    def test_missing(self, capsys, config_file):
        args = argparse.Namespace(target="foo")
        command = RmCommand(args, {"targets": {"bar": {}}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == ('"foo" is an unknown target\n' "Choices are: ['bar']\n")

    def test_remove_target(self, capsys, config_file):
        args = argparse.Namespace(target="foo")
        command = RmCommand(args, {"targets": {"bar": {}, "foo": {}}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == err == ""

        with open(config_file, "rt") as fp:
            new_config = json.load(fp)

        expected_config = {"targets": {"bar": {}}}
        assert new_config == expected_config



# -*- coding: utf-8 -*-

import mock
import pytest

from s4 import cli


# TODO: Should catch KeyboardExceptions and raise them again
class TestMain(object):
    @mock.patch("argparse.ArgumentParser.print_help")
    def test_no_arguments_prints_help(self, print_help):
        cli.main([])
        assert print_help.call_count == 1

    @pytest.mark.parametrize(["loglevel"], [("INFO",), ("DEBUG",)])
    @mock.patch("logging.basicConfig")
    def test_timestamps(self, basicConfig, loglevel):
        cli.main(["--timestamps", "--log-level", loglevel, "version"])
        assert basicConfig.call_args[1]["format"].startswith("%(asctime)s: ")

    @mock.patch("logging.basicConfig")
    def test_debug_loglevel(self, basicConfig):
        cli.main(["--log-level=DEBUG", "version"])
        assert basicConfig.call_args[1]["format"].startswith("%(levelname)s:%(module)s")
        assert basicConfig.call_args[1]["level"] == "DEBUG"

    @pytest.mark.parametrize("command", ["version", "v"])
    def test_version_command(self, capsys, command):
        cli.main([command])
        out, err = capsys.readouterr()
        assert out == "{}\n".format(cli.VERSION)

    @mock.patch("s4.cli.LsCommand")
    def test_ls_command(self, LsCommand):
        cli.main(["ls", "foo"])
        assert LsCommand.call_count == 1

    @pytest.mark.parametrize("command", ["daemon", "d"])
    @mock.patch("s4.cli.DaemonCommand")
    def test_daemon_command(self, DaemonCommand, command):
        cli.main([command])
        assert DaemonCommand.call_count == 1

    @pytest.mark.parametrize("command", ["sync", "s"])
    @mock.patch("s4.cli.SyncCommand")
    def test_sync_command(self, SyncCommand, command):
        cli.main([command, "foo"])
        assert SyncCommand.call_count == 1

    @pytest.mark.parametrize("command", ["edit", "e"])
    @mock.patch("s4.cli.EditCommand")
    def test_edit_command(self, EditCommand, command):
        cli.main([command, "foo"])
        assert EditCommand.call_count == 1

    @pytest.mark.parametrize("command", ["targets", "t"])
    @mock.patch("s4.cli.TargetsCommand")
    def test_targets_command(self, TargetsCommand, command):
        cli.main([command])
        assert TargetsCommand.call_count == 1

    @mock.patch("s4.cli.RmCommand")
    def test_rm_command(self, RmCommand):
        cli.main(["rm", "foo"])
        assert RmCommand.call_count == 1

    @pytest.mark.parametrize("command", ["add", "a"])
    @mock.patch("s4.cli.AddCommand")
    def test_add_command(self, AddCommand, command):
        cli.main([command])
        assert AddCommand.call_count == 1



# -*- coding: utf-8 -*-

import gzip
import io
import json
import os
import shutil
import tempfile

import filelock
import mock
import pytest

from s4.clients import SyncObject, local

from tests import utils


class TestTraverse(object):
    def setup_method(self):
        self.target_folder = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.target_folder)

    def test_non_existent_folder(self):
        assert list(local.traverse("/i/definetely/do/not/exist")) == []

    def test_empty_folder(self):
        assert list(local.traverse(self.target_folder)) == []

    def test_correct_output(self):
        items = [
            "baz/zoo",
            "foo",
            "bar.md",
            "baz/bar",
            ".index",
            "garbage~",
            "saw/.index",
            "SomeProject/.git/2aa58a13dcbca4b13a244dadf5536865ead5e1",
            "SomeProject/hello.py",
            "SomeProject/hello.pyo",
            "SomeProject/hello.pyc",
        ]
        for item in items:
            utils.write_local(os.path.join(self.target_folder, item))

        actual_output = list(
            local.traverse(
                self.target_folder,
                ignore_files={".index", ".idontexist", "*~", ".git", "*.py[co]"},
            )
        )

        expected_output = [
            "bar.md",
            "baz/bar",
            "baz/zoo",
            "foo",
            "SomeProject/hello.py",
        ]
        assert sorted(actual_output) == sorted(expected_output)

    def test_ignore_subfolder(self):
        items = [
            "dev/files/test.txt",
            "dev/venv/python2.7/pip-selfcheck.json",
            "dev/venv/python2.7/lib/file",
            "dev/.git/index",
            "dev/main.py",
        ]
        for item in items:
            utils.write_local(os.path.join(self.target_folder, item))
        actual_output = list(
            local.traverse(self.target_folder, ignore_files={"venv/", ".git/"})
        )
        expected_output = ["dev/main.py", "dev/files/test.txt"]
        assert sorted(actual_output) == sorted(expected_output)


class TestLocalSyncClient(object):
    def test_get_client_name(self, local_client):
        assert local_client.get_client_name() == "local"

    def test_repr(self):
        client = local.LocalSyncClient("/my/test/path")
        assert repr(client) == "LocalSyncClient</my/test/path>"

    def test_lock(self, local_client):
        local_client_2 = local.LocalSyncClient(local_client.path)
        local_client.lock(timeout=0.01)
        with pytest.raises(filelock.Timeout):
            local_client_2.lock(timeout=0.01)
        local_client.unlock()

        local_client.lock(timeout=0.01)
        local_client_2.unlock()

    def test_get_size_not_exists(self, local_client):
        assert local_client.get_size("idontexist") == 0

    def test_get_size_exists(self, local_client):
        utils.write_local(local_client.get_uri("foo"), "test data")
        assert local_client.get_size("foo") == 9

    def test_put_callback(self, local_client):
        mock_callback = mock.MagicMock()

        data = b"hello"
        local_client.put(
            key="solair/astora",
            sync_object=SyncObject(io.BytesIO(data), len(data), 7194),
            callback=mock_callback,
        )

        assert mock_callback.call_count == 1
        mock_callback.assert_called_with(5)

    def test_put_new(self, local_client):
        data = b"hi"
        local_client.put(
            key="foo/hello_world.txt",
            sync_object=SyncObject(io.BytesIO(data), len(data), 20000),
        )

        assert local_client.index["foo/hello_world.txt"]["remote_timestamp"] == 20000
        assert utils.get_local_contents(local_client, "foo/hello_world.txt") == data

    def test_get_uri(self):
        client = local.LocalSyncClient("/home/michael")
        assert client.get_uri() == "/home/michael/"
        assert client.get_uri("banana.txt") == "/home/michael/banana.txt"

    def test_put_existing(self, local_client):
        utils.set_local_index(local_client, {"doge.txt": {"local_timestamp": 1111111}})

        data = b"canis lupus familiaris"
        local_client.put(
            key="doge.txt", sync_object=SyncObject(io.BytesIO(data), len(data), 20000)
        )

        assert local_client.index["doge.txt"]["remote_timestamp"] == 20000
        assert local_client.index["doge.txt"]["local_timestamp"] == 1111111

        assert utils.get_local_contents(local_client, "doge.txt") == data

    def test_get_existing(self, local_client):
        data = b"blue green yellow"
        utils.set_local_contents(local_client, "whatup.md", data=data.decode("utf8"))
        sync_object = local_client.get("whatup.md")
        assert sync_object.fp.read() == data
        assert sync_object.total_size == len(data)

    def test_get_non_existent(self, local_client):
        assert local_client.get("idontexist.md") is None

    def test_delete_existing(self, local_client):
        target_file = os.path.join(local_client.path, "foo")
        utils.set_local_contents(local_client, "foo", 222222)

        assert os.path.exists(target_file) is True
        assert local_client.delete("foo") is True
        assert os.path.exists(target_file) is False

    def test_delete_non_existent(self, local_client):
        assert local_client.delete("idontexist.txt") is False

    def test_index_path(self):
        client = local.LocalSyncClient("/hello/from/the/magic/tavern")
        assert client.index_path() == "/hello/from/the/magic/tavern/.index"

    def test_load_index(self, local_client):
        data = {
            "foo": {"local_timestamp": 4000, "remote_timestamp": 4000},
            "bar/baz.txt": {"local_timestamp": 5000, "remote_timestamp": 5000},
        }
        utils.set_local_index(local_client, data)

        assert local_client.index == data

    def test_get_local_keys(self, local_client):
        utils.set_local_index(
            local_client, {}
        )  # .index file should not come up in results
        utils.set_local_contents(local_client, ".bashrc")
        utils.set_local_contents(local_client, "foo")
        utils.set_local_contents(local_client, "bar")

        actual_output = local_client.get_local_keys()
        expected_output = ["foo", "bar", ".bashrc"]
        assert sorted(expected_output) == sorted(actual_output)

    def test_get_index_keys(self, local_client):
        data = {
            "foo": {"local_timestamp": 4000, "remote_timestamp": 4000},
            "bar/baz.txt": {"local_timestamp": 5000, "remote_timestamp": 5000},
        }
        utils.set_local_index(local_client, data)

        actual_output = local_client.get_index_keys()
        expected_output = ["foo", "bar/baz.txt"]
        assert sorted(list(actual_output)) == sorted(expected_output)

    def test_all_keys(self, local_client):
        utils.set_local_contents(local_client, ".index")
        utils.set_local_contents(local_client, "foo")
        utils.set_local_contents(local_client, "bar/boo.md")
        data = {
            "foo": {"local_timestamp": 4000, "remote_timestamp": 4000},
            "bar/baz.txt": {"local_timestamp": 5000, "remote_timestamp": 5000},
        }
        utils.set_local_index(local_client, data)

        actual_output = local_client.get_all_keys()
        expected_output = ["foo", "bar/boo.md", "bar/baz.txt"]
        assert sorted(actual_output) == sorted(expected_output)

    def test_update_index_empty(self, local_client):
        utils.set_local_contents(local_client, "foo", 13371337)
        utils.set_local_contents(local_client, "bar", 50032003)

        local_client.update_index()
        # remote timestamp should not be included since it does not exist
        expected_output = {
            "foo": {"local_timestamp": 13371337, "remote_timestamp": None},
            "bar": {"local_timestamp": 50032003, "remote_timestamp": None},
        }
        assert local_client.index == expected_output

    def test_update_index_non_empty(self, local_client):
        utils.set_local_contents(local_client, "foo", 13371337)
        utils.set_local_contents(local_client, "bar", 50032003)

        utils.set_local_index(
            local_client,
            {
                "foo": {"local_timestamp": 4000, "remote_timestamp": 4000},
                "bar": {"local_timestamp": 5000, "remote_timestamp": 5000},
                "baz": {"local_timestamp": 5000, "remote_timestamp": 5000},
            },
        )

        local_client.update_index()
        expected_output = {
            "foo": {"local_timestamp": 13371337, "remote_timestamp": 4000},
            "bar": {"local_timestamp": 50032003, "remote_timestamp": 5000},
            "baz": {"local_timestamp": None, "remote_timestamp": 5000},
        }
        assert local_client.index == expected_output

    def test_get_real_local_timestamp(self, local_client):
        utils.set_local_contents(local_client, "atcg", 2323230)

        assert local_client.get_real_local_timestamp("atcg") == 2323230
        assert local_client.get_real_local_timestamp("dontexist") is None

    def test_get_all_real_local_timestamps(self, local_client):
        utils.set_local_contents(local_client, "red", 2323230)
        utils.set_local_contents(local_client, "blue", 80808008)

        expected_output = {"red": 2323230, "blue": 80808008}
        actual_output = local_client.get_all_real_local_timestamps()
        assert actual_output == expected_output

    def test_get_index_timestamps(self, local_client):
        utils.set_local_index(
            local_client,
            {
                "foo": {"local_timestamp": 4000, "remote_timestamp": 32000},
                "bar": {"local_timestamp": 9999999},
            },
        )
        assert local_client.get_index_local_timestamp("foo") == 4000
        assert local_client.get_index_local_timestamp("dontexist") is None
        assert local_client.get_index_local_timestamp("bar") == 9999999

        assert local_client.get_remote_timestamp("foo") == 32000
        assert local_client.get_remote_timestamp("dontexist") is None
        assert local_client.get_remote_timestamp("bar") is None

    def test_set_index_local_timestamp(self, local_client):
        utils.set_local_index(
            local_client,
            {"foo": {"local_timestamp": 100000, "remote_timestamp": 2222222222}},
        )
        local_client.set_index_local_timestamp("foo", 123456)
        assert local_client.get_index_local_timestamp("foo") == 123456

    def test_set_index_local_timestamp_non_existent(self, local_client):
        local_client.set_index_local_timestamp("foo", 222222)
        assert local_client.get_index_local_timestamp("foo") == 222222

    def test_get_all_index_local_timestamps(self, local_client):
        utils.set_local_index(
            local_client,
            {"frap": {"local_timestamp": 4000}, "brap": {"local_timestamp": 9999999}},
        )

        expected_output = {"frap": 4000, "brap": 9999999}
        actual_output = local_client.get_all_index_local_timestamps()
        assert actual_output == expected_output

    def test_get_all_remote_timestamps(self, local_client):
        utils.set_local_index(
            local_client,
            {"frap": {"remote_timestamp": 4000}, "brap": {"remote_timestamp": 9999999}},
        )

        expected_output = {"frap": 4000, "brap": 9999999}
        actual_output = local_client.get_all_remote_timestamps()
        assert actual_output == expected_output

    def test_interrupted_put(self, local_client):
        utils.set_local_contents(local_client, "keychain", data="iamsomedatathatexists")

        sync_object = SyncObject(utils.InterruptedBytesIO(), 900000, 3000)

        with pytest.raises(ValueError):
            local_client.put("keychain", sync_object)

        result = local_client.get("keychain")
        assert result.fp.read() == b"iamsomedatathatexists"

    @pytest.mark.parametrize(
        ["compressed", "method"], [(True, gzip.open), (False, open)]
    )
    def test_flush_index(self, compressed, method, local_client):
        target_index = {"foo": {"local_timestamp": 4000, "remote_timestamp": 6000}}

        local_client.index = target_index
        local_client.flush_index(compressed=compressed)

        with method(local_client.index_path(), "rt") as fp:
            index = json.load(fp)

        assert index == target_index

    def test_interrupted_flush_index(self, local_client):
        target_index = {"red": {"local_timestamp": 4000, "remote_timestamp": 4000}}

        utils.set_local_index(local_client, target_index)

        local_client.index = {"invalid_data": []}

        with mock.patch("json.dump") as json_dump:
            json_dump.side_effect = ValueError("something went wrong")
            with pytest.raises(ValueError):
                local_client.flush_index(compressed=False)

        with open(local_client.index_path(), "rt") as fp:
            index = json.load(fp)

        assert index == target_index

    def test_ignore_files(self, local_client):
        utils.set_local_contents(
            local_client, ".syncignore", timestamp=3200, data=("*.zip\n" "foo*\n")
        )
        local_client.reload_ignore_files()

        utils.set_local_index(
            local_client,
            {"pony.tar": {"local_timestamp": 4000, "remote_timestamp": 3000}},
        )

        utils.set_local_contents(local_client, "test.zip")
        utils.set_local_contents(local_client, "foobar")
        utils.set_local_contents(local_client, "foo")
        utils.set_local_contents(local_client, "pony.tar", timestamp=8000)

        local_client.update_index()

        expected_index = {
            ".syncignore": {"local_timestamp": 3200, "remote_timestamp": None},
            "pony.tar": {"local_timestamp": 8000, "remote_timestamp": 3000},
        }
        assert local_client.index == expected_index



#! -*- encoding: utf-8 -*-
import argparse

import mock

from s4.clients import SyncState
from s4.commands.sync_command import (
    SyncCommand,
    display_progress_bar,
    handle_conflict,
    hide_progress_bar,
    update_progress_bar,
)
from s4.resolution import Resolution
from s4.sync import SyncWorker

from tests.utils import create_logger, set_local_contents


@mock.patch("s4.utils.get_input")
class TestHandleConflict(object):
    def test_first_choice(self, get_input, s3_client, local_client):
        get_input.return_value = "1"

        action_1 = SyncState(SyncState.UPDATED, 1111, 2222)
        action_2 = SyncState(SyncState.DELETED, 3333, 4444)

        result = handle_conflict("movie", action_1, s3_client, action_2, local_client)
        assert result.action == Resolution.UPDATE

    def test_second_choice(self, get_input, s3_client, local_client):
        get_input.return_value = "2"

        action_1 = SyncState(SyncState.UPDATED, 1111, 2222)
        action_2 = SyncState(SyncState.DELETED, 3333, 4444)

        result = handle_conflict("movie", action_1, s3_client, action_2, local_client)
        assert result.action == Resolution.DELETE

    def test_skip(self, get_input, s3_client, local_client):
        get_input.return_value = "X"

        action_1 = SyncState(SyncState.UPDATED, 1111, 2222)
        action_2 = SyncState(SyncState.DELETED, 3333, 4444)

        result = handle_conflict("movie", action_1, s3_client, action_2, local_client)
        assert result is None

    @mock.patch("s4.commands.sync_command.show_diff")
    def test_diff(self, show_diff, get_input, s3_client, local_client):
        get_input.side_effect = ["d", "X"]

        action_1 = SyncState(SyncState.UPDATED, 1111, 2222)
        action_2 = SyncState(SyncState.DELETED, 3333, 4444)

        result = handle_conflict("movie", action_1, s3_client, action_2, local_client)
        assert result is None
        assert show_diff.call_count == 1
        show_diff.assert_called_with(s3_client, local_client, "movie")


def test_progressbar_smoketest(s3_client, local_client):
    # Just test that nothing blows up
    set_local_contents(
        local_client, "history.txt", data="a long long time ago", timestamp=5000
    )

    worker = SyncWorker(
        s3_client,
        local_client,
        start_callback=display_progress_bar,
        update_callback=update_progress_bar,
        complete_callback=hide_progress_bar,
    )
    worker.sync()


@mock.patch("s4.sync.SyncWorker")
class TestSyncCommand(object):
    def test_no_targets(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        command = SyncCommand(args, {"targets": {}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == err == ""
        assert SyncWorker.call_count == 0

    def test_wrong_target(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=["foo", "bar"], conflicts=None, dry_run=False)
        command = SyncCommand(args, {"targets": {"baz": {}}}, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            "\"bar\" is an unknown target. Choices are: ['baz']\n"
            "\"foo\" is an unknown target. Choices are: ['baz']\n"
        )
        assert SyncWorker.call_count == 0

    def test_sync_error(self, SyncWorker, capsys):
        args = argparse.Namespace(
            targets=None, conflicts=None, dry_run=False, log_level="INFO"
        )
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mike/docs",
                    "s3_uri": "s3://foobar/docs",
                    "aws_access_key_id": "3223323",
                    "aws_secret_access_key": "23#@423#@",
                    "region_name": "us-east-1",
                },
                "bar": {
                    "local_folder": "/home/mike/barmil",
                    "s3_uri": "s3://foobar/barrel",
                    "aws_access_key_id": "3223",
                    "aws_secret_access_key": "23#eWEa@423#@",
                    "region_name": "us-west-2",
                },
            }
        }
        SyncWorker.side_effect = ValueError("something bad happened")

        command = SyncCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 2
        assert out == ""
        assert err == (
            "There was an error syncing 'bar':\nsomething bad happened\n"
            "There was an error syncing 'foo':\nsomething bad happened\n"
        )

    def test_sync_error_debug(self, SyncWorker, capsys):
        args = argparse.Namespace(
            targets=None, conflicts=None, dry_run=False, log_level="DEBUG"
        )
        config = {
            "targets": {
                "bar": {
                    "local_folder": "/home/mike/barmil",
                    "s3_uri": "s3://foobar/barrel",
                    "aws_access_key_id": "3223",
                    "aws_secret_access_key": "23#eWEa@423#@",
                    "region_name": "us-west-2",
                }
            }
        }
        SyncWorker.side_effect = ValueError("something bad happened")

        command = SyncCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 1
        assert out == ""
        assert err.split("\n")[:2] == [
            "something bad happened",
            "Traceback (most recent call last):",
        ]

    def test_keyboard_interrupt(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mike/docs",
                    "s3_uri": "s3://foobar/docs",
                    "aws_access_key_id": "3223323",
                    "aws_secret_access_key": "23#@423#@",
                    "region_name": "us-east-1",
                },
                "bar": {
                    "local_folder": "/home/mike/barmil",
                    "s3_uri": "s3://foobar/barrel",
                    "aws_access_key_id": "3223",
                    "aws_secret_access_key": "23#eWEa@423#@",
                    "region_name": "us-west-2",
                },
            }
        }
        SyncWorker.side_effect = KeyboardInterrupt

        command = SyncCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert SyncWorker.call_count == 1
        assert out == ""
        assert err == ("Quitting due to Keyboard Interrupt...\n")

    def test_all_targets(self, SyncWorker, capsys):
        args = argparse.Namespace(targets=None, conflicts=None, dry_run=False)
        config = {
            "targets": {
                "foo": {
                    "local_folder": "/home/mike/docs",
                    "s3_uri": "s3://foobar/docs",
                    "aws_access_key_id": "3223323",
                    "aws_secret_access_key": "23#@423#@",
                    "region_name": "us-east-1",
                },
                "bar": {
                    "local_folder": "/home/mike/barmil",
                    "s3_uri": "s3://foobar/barrel",
                    "aws_access_key_id": "3223",
                    "aws_secret_access_key": "23#eWEa@423#@",
                    "region_name": "us-west-2",
                },
            }
        }

        command = SyncCommand(args, config, create_logger())
        command.run()

        out, err = capsys.readouterr()
        assert out == ""
        assert err == (
            "Syncing bar [/home/mike/barmil/ <=> s3://foobar/barrel/]\n"
            "Syncing foo [/home/mike/docs/ <=> s3://foobar/docs/]\n"
        )
        assert SyncWorker.call_count == 2
